/*
 *	detar.c
 *
 *  Version 1.0
 *
 *	Unix Tape Archive (TAR) Extractor for MSDOS
 *
 *	Public Domain January 1988, Steve Sampson
 *	Based on the public domain TAR program by John Gilmore
 *
 *  Compiled with ECO-C88 V3.21
 *
 *  version 1.1
 *
 *  Tim Riker - added directory handling while detarring
 *              fixed various bugs, now stops at end of tar
 *              added setvbuf calls (as every good programmer should)
 *              ANSIized function defs (sorry, all you with old compilers!)
 *
 *  Public Domain November 1991, Tim Riker
 *
 *  compiled with Microsoft QuickC 2.51 (Workbench is slow...)
 */

/* uncomment this line to use our ctime */
/* #define NO_CTIME */

/* Includes */

#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <direct.h>
#include <errno.h>

#ifndef NO_CTIME
#include <time.h>
#endif

#include "detar.h"

/* Defines */

#define	isodigit(c)	( ((c) >= '0') && ((c) <= '7') )
#define BUFSIZE 30720

/* Globals */

FILE *fdi;
FILE *fdo;

union record *head;         /* the archive file header */
struct stat hstat;          /* the decoded stat info   */
struct
{
	int	year;
	int	month;
	int	day;
	int	hour;
	int	min;
	int	sec;
} tm;

long s_p_min;
long s_p_hour;
long s_p_day;
long s_p_year;

int ugswidth = 11;
int dir_only = 0;
int days_per_month[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

char tarname[128];

/* Forward References, Prototypes */

#ifdef NO_CTIME
char    *ctime(long *);
#endif
int main(int argc, char *argv[]);
long from_oct(int digs, char *where);
char *convert(char *unixname);
int dot_convert(char* convname, char *unixname);
void cv_time(long time);
void decode_header(union record *header, struct stat *st, int *stdp);
void dir_tar(void);
void de_tar(void);
void demode(unsigned int mode, char *string);

#ifdef NO_CTIME
char    *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"},
	*days[] = {"Thu", "Fri", "Sat", "Sun", "Mon", "Tue", "Wed"},
	t_buf[26];

/*
 *  Borrowed from comp.os.minix
 */

char *ctime(long *time)
{
	long    tt;

	s_p_min  = 60L;
	s_p_hour = 60L * 60L;
	s_p_day  = 60L * 60L * 24L;
	s_p_year = 60L * 60L * 24L * 365L;

	tt = *time;
	cv_time(tt);

	sprintf(t_buf, "%s %s %02d %02d:%02d:%02d %04d\n",
	   days[(tt / s_p_day) % 7], months[tm.month], tm.day, tm.hour,
		 tm.min, tm.sec, tm.year);

	return(t_buf);
}
#endif


/* Main Program */

int main(int argc, char *argv[])
{
	char *p;

	/*
	 *	The operator can ask for directory only
	 *	or de-tar only.
	 */

	argv++;

	if (argc == 3 && **argv == 'd')
	{
		dir_only++;
		argc--;
		argv++;
	}

	if (argc != 2)
	{
		fprintf(stderr, "Usage: detar [d] filename[.tar]\n");
		fprintf(stderr, "Where d means directory only\n");
		exit(1);
	}

	for (p = *argv; *p; p++)
		*p = (char)toupper(*p);

	/* parse the many possibilities, this stuff not needed with MSC */

	if ((p = strrchr(*argv, '.')) == (char *)NULL)
		goto add;
	else if (strcmp(p, ".TAR") == 0)
		strcpy(tarname, *argv);
	else if (*(p+1) == '/' || *(p+1) == '\\')
	{
add:        strcpy(tarname, *argv);
		strcat(tarname, ".TAR");
	}
	else
	{
		fprintf(stderr, "File '%s' not a TAR archive\n", *argv);
		exit(1);
	}

	if ((fdi = fopen(tarname, "rb")) == (FILE *)NULL)
	{
		fprintf(stderr, "Tar File '%s' Not Found\n", tarname);
		exit(1);
	}
	setvbuf(fdi, NULL, _IOFBF, BUFSIZE);

	head = (union record *)malloc(512);

	if (dir_only)
		dir_tar();
	else
		de_tar();

	return(0);
}


/*
 *	Produce a directory of the files contained on the TAR
 */

void dir_tar(void)
{
	char	modes[11], *timestamp;
	char	uform[11], gform[11];
	char	*user, *group, size[24];
	long	longie;
	int	n, pad, header_std;

	for (;;)
	{
		if ((n = fread((char *)head, 1, 512, fdi)) != 512)
		{
			fclose(fdi);
			free(head);
			return;
		}

		decode_header(head, &hstat, &header_std);

		if (head->header.name[0] == '\0')
			return;

		/* File type and modes */

		modes[0] = '?';

		switch (head->header.linkflag)
		{
		case LF_NORMAL:
		case LF_OLDNORMAL:
		case LF_LINK:
			modes[0] = '-'; 
			if ('/' == head->header.name[strlen(head->header.name)-1])
				modes[0] = 'd';
			break;
		case LF_DIR:
			modes[0] = 'd';
			break;
		case LF_SYMLINK:
			modes[0] = 'l';
			break;
		case LF_BLK:
			modes[0] = 'b';
			break;
		case LF_CHR:
			modes[0] = 'c';
			break;
		case LF_FIFO:
			modes[0] = 'f';
			break;
		case LF_CONTIG:
			modes[0] = '=';
			break;
		}

		demode(hstat.st_mode, modes+1);

		/* Timestamp */

		longie = hstat.st_mtime;
		timestamp = ctime(&longie);
		timestamp[16] = '\0';
		timestamp[24] = '\0';

		/* User and group names */

		if (*head->header.uname && header_std)
		{
			user  = head->header.uname;
		}
		else
		{
			user = uform;
			sprintf(uform, "%d", hstat.st_uid);
		}

		if (*head->header.gname && header_std)
		{
			group = head->header.gname;
		}
		else
		{
			group = gform;
			sprintf(gform, "%d", hstat.st_gid);
		}

		/* Format the file size or major/minor device numbers */

		switch (head->header.linkflag)
		{
		case LF_CHR:
		case LF_BLK:
			sprintf(size, "%d, %d",
			   (int)from_oct(8, head->header.devmajor),
			   (int)from_oct(8, head->header.devminor));
			break;
		default:
			sprintf(size, "%ld", hstat.st_size);
		}

		/* Figure out padding and print the whole line. */ 
		pad = strlen(user) + strlen(group) + strlen(size) + 1;

		if (pad > ugswidth)
			ugswidth = pad;

		printf("%s %s/%s %*s%s %s %s %.*s",
			modes,
			user,
			group,
			ugswidth - pad,
			"",
			size,
			timestamp+4, timestamp+20,
			sizeof(head->header.name),
			head->header.name);

		switch (head->header.linkflag)
		{
		case LF_SYMLINK:
			printf(" -> %s\n", head->header.linkname);
			break;

		case LF_LINK:
			printf(" link to %s\n", head->header.linkname);
			break;

		default:
			printf(" unknown file type '%c'\n", head->header.linkflag);
			break;

		case LF_OLDNORMAL:
		case LF_NORMAL:
		case LF_CHR:
		case LF_BLK:
		case LF_DIR:
		case LF_FIFO:
		case LF_CONTIG:
			putc('\n', stdout);
			break;
		}

		/*
		 *	Seek to next file
		 */

		fseek(fdi, hstat.st_size, SEEK_CUR);

		/*
		 *	File starts on 512 byte boundary
		 */

		if (hstat.st_size % 512L)
			fseek(fdi, 512L - (hstat.st_size % 512L), 1);
	}
}


/*
 *	Extract the files from the TAR archive
 *
 */

void de_tar(void)
{
	long	size;
	int	header_std;
	char	*workfile;
	char	c;

	for (;;)
	{
		if ( fread((char *)head, 1, 512, fdi) != 512)
		{
			fclose(fdi);
			free(head);
			return;
		}

		decode_header(head, &hstat, &header_std);

		if (head->header.name[0] == '\0')
			return;

		workfile = convert(head->header.name);

		if ('/' == head->header.name[strlen(head->header.name)-1])
		/* if (head->header.linkflag == LF_DIR) */
		{
			workfile[strlen(workfile) - 1] = '\0';
			printf("Create dir: %s\n", workfile);
			mkdir(workfile);
		}
		else
		{
			printf("Extracting: %s\n", workfile);

			fdo = fopen(workfile, "wb");
			setvbuf(fdo, NULL, _IOFBF, BUFSIZE);
			size = hstat.st_size;

			if (size > (long)0)
			{
				while (size > (long)0)
				{
					c = (char)getc(fdi);
#ifdef CR_LF_CHANGE
					/*
					 *  Convert Linefeed to MSDOS
					 *  Carriage Return, Linefeed pair
					 *
					 */

					if (c == 0x0A)
						putc('\n', fdo);
					else
#endif
						putc(c, fdo);

					size--;
				}
				if (size == 0L)
				{

				   /*
					*   Get to next 512 byte boundary
					*/

				   if (hstat.st_size % 512L)
					   fseek(fdi, 512L - (hstat.st_size % 512L), 1);
				}
				fclose(fdo);
			}
		}
	}
}


/* 
 *	Break down the header info into stat info
 *
 *	Set "*stdp" to != 0 or == 0 depending whether
 *	header record is "Unix Standard" or "old" tar format.
 *
 */

void decode_header(union record *header, struct stat *st, int *stdp)
{
	st->st_mode  = (int)from_oct( 8, header->header.mode);
	st->st_mtime = from_oct(12, header->header.mtime);
	st->st_size  = from_oct(12, header->header.size);
	
	if (0 == strcmp(header->header.magic, TMAGIC))
	{

		/* Unix Standard tar archive */

		*stdp = 1;
		st->st_dev = 0;

	}
	else
	{

		/* Old fashioned tar archive */

		*stdp = 0;
		st->st_uid = (int)from_oct(8, header->header.uid);
		st->st_gid = (int)from_oct(8, header->header.gid);
		st->st_dev = 0;
	}
}


/*
 *	Decode the mode string from a stat entry into a 9-char string
 */

void demode(unsigned int mode, char *string)
{
	unsigned int mask;
	char *rwx = "rwxrwxrwx";

	for (mask = 0400; mask != 0; mask >>= 1)
	{
		if (mode & mask)
			*string++ = *rwx++;
		else
		{
			*string++ = '-';
			rwx++;
		}
	}

	if (mode & S_ISUID)
		if (string[-7] == 'x')
			string[-7] = 's';
		else
			string[-7] = 'S';

	if (mode & S_ISGID)
		if (string[-4] == 'x')
			string[-4] = 's';
		else
			string[-4] = 'S';

	if (mode & S_ISVTX)
		if (string[-1] == 'x')
			string[-1] = 't';
		else
			string[-1] = 'T';

	*string = '\0';
}


/*
 *	Quick and dirty octal conversion.
 *
 *	Result is -1 if the field is invalid (all blank, or nonoctal).
 */

long from_oct(int digs, char *where)
{
	long   value;

	while (isspace(*where))
	{       /* Skip spaces */
		where++;
		if (--digs <= 0)
			return -1;		/* All blank field */
	}

	value = 0;
	while (digs > 0 && isodigit(*where))
	{  /* Scan til nonoctal */
		value = (value << 3) | (*where++ - '0');
		--digs;
	}

	if (digs > 0 && *where && !isspace(*where))
		return -1;			/* Ended on non-space/nul */

	return value;
}


void cv_time(long time)
{
	tm.year = 0;
	tm.month = 0;
	tm.day = 1;
	tm.hour = 0;
	tm.min = 0;
	tm.sec = 0;

	while (time >= s_p_year)
	{
		if (((tm.year + 2) % 4) == 0)
			time -= s_p_day;

		tm.year += 1;
		time -= s_p_year;
	}

	if (((tm.year + 2) % 4) == 0)
		days_per_month[1]++;

	tm.year += 1970;

	while ( time >= (days_per_month[tm.month] * s_p_day))
		time -= days_per_month[tm.month++] * s_p_day;

	while (time >= s_p_day)
	{
		time -= s_p_day;
		tm.day++;
	}

	while (time >= s_p_hour)
	{
		time -= s_p_hour;
		tm.hour++;
	}

	while (time >= s_p_min)
	{
		time -= s_p_min;
		tm.min++;
	}

	tm.sec = (int)time;
}

char *convert(char *unixname)
{
	static char convname[NAMSIZ];
	char tempname[NAMSIZ];
	size_t unix, conv, temp;

	conv = temp = 0;
	memset(convname, 0, NAMSIZ);
	memset(tempname, 0, NAMSIZ);

	for (unix = 0; unix < strlen(unixname); unix++)
	{
		switch (unixname[unix])
		{
			case '/':
			{
				tempname[temp++] = '\0';
				conv += dot_convert(&convname[conv], &tempname[0]);
				convname[conv++] = '\\';
				memset(tempname, 0, NAMSIZ);
				temp = 0;
				break;
			}
			default:
				tempname[temp++] = unixname[unix];
		}
	}
	conv += dot_convert(&convname[conv], &tempname[0]);
	return(convname);
}

int dot_convert(char* convname, char *unixname)
{
	char   *p;
	int    have_dot;

	strcpy(convname, unixname);

	/* make a basic MSDOS filename */

	if ((p = strchr(convname, '.')) != NULL)
	{
		if (strlen(p) > 4)
			p[4] = '\0';

		if ((p - convname) > 8)
		{
			convname[8] = '\0';
			strcat(convname, p);
		}
	}
	else
		convname[8] = '\0';

	/* Now look for two dots in filename, only one allowed in MSDOS */
	/* substitute a dash for second dot */

	for (have_dot = 0, p = convname; *p; p++)
	{
		if (*p == '.')
		{
			if (have_dot)
				*p = '-';
			else
				have_dot = 1;
		}
	}

	return(strlen(convname));
}

/* EOF */
 
